import os

import boto3

from cloud_governance.common.clouds.aws.utils.common_methods import get_boto3_client
from cloud_governance.common.clouds.aws.utils.utils import Utils
from cloud_governance.common.logger.init_logger import logger
from datetime import datetime, timezone


class IAMOperations:

    ACCESS_KEY_LABEL_MAP = {"access key 1": 0, "access key 2": 1}

    def __init__(self, iam_client=None):
        self.iam_client = iam_client if iam_client else get_boto3_client('iam')
        self.utils = Utils()
        self.__sts_client = boto3.client('sts')

    @property
    def get_iam_client(self):
        return self.iam_client

    def get_user_tags(self, username: str):
        """
        This method return tags from the iam resources
        @param username:
        @return:
        """
        try:
            user = self.iam_client.get_user(UserName=username)['User']
            if user.get('Tags'):
                return user.get('Tags')
            else:
                return []
        except:
            return []

    def get_roles(self):
        """
        This method returns all roles
        @return:
        """
        return self.utils.get_details_resource_list(func_name=self.iam_client.list_roles, input_tag='Roles',
                                                    check_tag='Marker')

    def get_users(self):
        """
        This method returns all users
        @return:
        """
        return self.utils.get_details_resource_list(self.iam_client.list_users, input_tag='Users', check_tag='Marker')

    def get_account_alias_cloud_name(self):
        """
        This method returns the aws account alias and cloud name
        @return:
        """
        try:
            account_alias = self.iam_client.list_account_aliases()['AccountAliases']
            if account_alias:
                return account_alias[0].upper(), 'AwsCloud'.upper()
        except:
            return os.environ.get('account', '').upper(), 'AwsCloud'.upper()

    def get_iam_users_list(self):
        """
        This method return the IAM users list
        :return:
        """
        iam_users = []
        users = self.get_users()
        for user in users:
            iam_users.append(user.get('UserName'))
        return iam_users

    def get_aws_account_id_name(self):
        """
        This method returns the aws account_id
        :return:
        """
        response = self.__sts_client.get_caller_identity()
        account_id = response['Account']
        return account_id

    def get_role(self, role_name: str):
        """
        This method returns the iam role data
        :param role_name:
        :return:
        """
        role_data = {}
        try:
            role_data = self.iam_client.get_role(RoleName=role_name).get('Role')
        except Exception as err:
            logger.error(err)
        return role_data

    def list_inline_role_policies(self, role_name: str):
        """
        This method returns the iam role inline policies
        :param role_name:
        :return:
        """
        role_policies = []
        try:
            role_policies = self.iam_client.list_role_policies(RoleName=role_name).get('PolicyNames', [])
        except Exception as err:
            logger.error(err)
        return role_policies

    def list_attached_role_policies(self, role_name: str):
        """
        This method returns the iam role attached policies
        :param role_name:
        :return:
        """
        attached_policies = []
        try:
            attached_policies = self.iam_client.list_attached_role_policies(RoleName=role_name).get('AttachedPolicies',
                                                                                                    [])
        except Exception as err:
            logger.error(err)
        return attached_policies

    def delete_role(self, role_name: str):
        """
        This method deletes the iam role
        :param role_name:
        :return:
        """
        try:
            self.iam_client.delete_role(RoleName=role_name)
            return True
        except Exception as err:
            raise err

    def tag_role(self, role_name: str, tags: list):
        """
        This method tags the iam role
        :param role_name:
        :param tags:
        :return:
        """
        try:
            self.iam_client.tag_role(RoleName=role_name, Tags=tags)
            return True
        except Exception as err:
            raise err

    def untag_role(self, role_name: str, tags: list):
        """
        This method untags the iam role
        :param role_name:
        :param tags:
        :return:
        """
        try:
            self.iam_client.untag_role(RoleName=role_name,
                                       TagKeys=[key for tag in tags for key, _ in tag.items() if key == 'Key'])
            return True
        except Exception as err:
            raise err

    def tag_user(self, user_name: str, tags: list):
        """
        This method tags the IAM user.
        :param user_name: The name of the IAM user to tag.
        :param tags: A list of tags to associate with the user.
        :return: True if tagging is successful, otherwise raises an exception.
        """
        try:
            self.iam_client.tag_user(UserName=user_name, Tags=tags)
            return True
        except Exception as err:
            raise err

    def get_iam_users_access_keys(self):
        """
        Retrieves IAM users and summarizes:
            - Access key status (active/inactive)
            - Access key age in days
            - Access key last used in days (or "N/A" if never used)
            - Tags (as a list of dictionaries)
            - Most recent key usage: last_activity_days
            - IAM client region (global context, since IAM is non-regional)
            - IAM user unique ID: ResourceId

        Returns:
            dict: {
                "username": {
                    "Access key 1": [status, age_days, last_used_days],
                    "Access key 2": [...],
                    "last_activity_days": int or "N/A",
                    "tags": [{"Key": "tag_key", "Value": "tag_value"}, ...],
                    "region": "us-east-1",
                    "ResourceId": "AIDAEXAMPLEUSERID"
                },
                ...
            }
        """
        result = {}
        now = datetime.now(timezone.utc)
        region_name = self.iam_client.meta.region_name or "global"

        paginator = self.iam_client.get_paginator('list_users')
        for page in paginator.paginate():
            for user in page['Users']:
                username = user['UserName']
                result[username] = {}
                # Access keys
                access_keys = self.iam_client.list_access_keys(UserName=username)['AccessKeyMetadata']
                for idx, key in enumerate(access_keys, start=1):
                    label = f"Access key {idx}"
                    status = key['Status'].lower()
                    age_days = (now - key['CreateDate']).days

                    # Get access key last used
                    try:
                        response = self.iam_client.get_access_key_last_used(AccessKeyId=key['AccessKeyId'])
                        last_used_date = response.get('AccessKeyLastUsed', {}).get('LastUsedDate')
                        if last_used_date:
                            last_used_days = (now - last_used_date).days
                        else:
                            # when no last used date is available, get the age of the key
                            last_used_days = age_days
                    except Exception:
                        # Set None in case of error to avoid deletion, not "N/A" because it will take age time
                        logger.error(f"Failed to get last used date for access key")
                        last_used_days = None

                    result[username][label] = {'label': label, 'status': status, 'age_days': age_days, 'last_activity_days': last_used_days}

                # Tags as list of dicts
                try:
                    tag_response = self.iam_client.list_user_tags(UserName=username)
                    tags = tag_response.get('Tags', [])
                except Exception:
                    tags = []

                result[username]["tags"] = tags
                result[username]["region"] = region_name
                result[username]["ResourceId"] = user.get('UserId')  # <-- Unique ID

        return result

    def has_active_access_keys(self, username: str, access_key_label: str = None) -> bool:
        """
        Checks if the given IAM user has any active access keys.
        Optionally filters by access key label ("Access Key 1" or "Access Key 2").

        Args:
            username (str): IAM user name
            access_key_label (str): Label to filter access keys ("Access Key 1"/"Access Key 2")

        Returns:
            bool: True if any access key is active (and matches the label if provided), False otherwise
        """
        try:
            keys = self.iam_client.list_access_keys(UserName=username)['AccessKeyMetadata']
        except Exception as e:
            logger.error(f"Failed to list access keys for user '{username}': {e}")
            return False

        # Sort keys by CreateDate ascending (oldest first)
        keys.sort(key=lambda k: k['CreateDate'])

        if access_key_label:
            idx = self.ACCESS_KEY_LABEL_MAP.get(access_key_label.lower())
            if idx is None or idx >= len(keys):
                return False
            return keys[idx].get('Status') == 'Active'

        return any(k.get('Status') == 'Active' for k in keys)

    def deactivate_user_access_key(self, username: str, **kwargs):
        """
        Deactivates the specified access key for the given IAM user.

        Args:
            username (str): IAM user name
            access_key_label (str): Access Key 1 or Access Key 2 (case-insensitive)
        """
        access_key_label = kwargs.get('access_key_label', '').lower()
        if not access_key_label:
            logger.warning("No access key label provided for deactivation.")
            return

        try:
            access_keys = self.iam_client.list_access_keys(UserName=username)['AccessKeyMetadata']
        except Exception as e:
            logger.error(f"Failed to list access keys for user '{username}': {e}")
            return

        # Sort keys by CreateDate ascending (oldest first) for consistent indexing
        access_keys.sort(key=lambda k: k['CreateDate'])

        idx = self.ACCESS_KEY_LABEL_MAP.get(access_key_label)
        if idx is None or idx >= len(access_keys):
            logger.warning(f"Access key label '{access_key_label}' not found for user '{username}'")
            return

        key_to_deactivate = access_keys[idx]
        access_key_id = key_to_deactivate['AccessKeyId']
        current_status = key_to_deactivate['Status'].lower()

        if current_status == 'active':
            try:
                self.iam_client.update_access_key(
                    UserName=username,
                    AccessKeyId=access_key_id,
                    Status='Inactive'
                )
                logger.info(f"Access key '{access_key_id}' deactivated for user '{username}'")
            except Exception as e:
                logger.error(f"Failed to deactivate access key '{access_key_id}' for user '{username}': {e}")
        else:
            logger.info(f"Access key '{access_key_id}' is already inactive for user '{username}'")

        logger.info(f"Access key deactivation processed for user '{username}'.")



from cloud_governance.common.logger.init_logger import logger
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp
from cloud_governance.policy.policy_operations.ibm.tagging.tagging_operations import TaggingOperations


class TagVM(TaggingOperations):
    """
    This class tags IBM virtual machines
    """

    def __init__(self):
        super().__init__()

    def get_virtual_machine_username(self, vm_id: str):
        """
        This method returns the virtual machine username from the billing order lists
        @param vm_id:
        @return:
        """
        vm_data = self._classic_operations.get_virtual_machine_data(vm_id=str(vm_id))
        return vm_data.get('billingItem').get('orderItem').get('order').get('userRecord').get(
            'username'), f'{vm_data.get("hostname")}.{vm_data.get("domain")}'

    def tag_update_virtual_machine(self, user_tags: list, vm_tags: list, vm_id: str, vm_name: str):
        """
        this method updates th etags of the virtual machine
        @param user_tags:
        @param vm_tags:
        @param vm_id:
        @param vm_name:
        @return:
        """
        add_vm_tags = []
        if self._tag_custom:
            user_tags.extend(self._tag_custom)
        add_vm_tags.extend(self._filter_common_tags(user_tags=user_tags, resource_tags=vm_tags))
        if add_vm_tags:
            tags = add_vm_tags
            if self._tag_operation == 'update':
                add_vm_tags.extend(vm_tags)
                add_vm_tags = list(set(add_vm_tags))
                try:
                    response = self.softlayer_operation(softlayer_name='SoftLayer_Virtual_Guest',
                                                        softlayer_method='setTags', tags=','.join(add_vm_tags),
                                                        resource_id=vm_id)
                    if response:
                        logger.info(f'Tags are added to the vm: {vm_id} - {vm_name} : count: {len(tags)} : {tags}')
                    else:
                        logger.info(f'Tags are not added to the vm: {vm_id} - {vm_name}, something might fail')
                except Exception as err:
                    logger.info(f'{err}')
        return add_vm_tags

    def tag_remove_virtual_machine(self, user_tags: list, vm_tags: list, vm_id: str, vm_name: str):
        """
        This method returns the tags of the virtual machine
        @param user_tags:
        @param vm_tags:
        @param vm_id:
        @param vm_name:
        @return:
        """
        remove_vm_tags = self._filter_remove_tags(user_tags=user_tags, resource_tags=vm_tags)
        if self._tag_remove_name:
            remove_vm_tags = [self._tag_remove_name]
        if remove_vm_tags:
            try:
                response = self.softlayer_operation(softlayer_name='SoftLayer_Virtual_Guest',
                                                    softlayer_method='removeTags', tags=','.join(remove_vm_tags),
                                                    resource_id=vm_id)
                if response:
                    logger.info(
                        f'Tags are Removed to the vm: {vm_id} - {vm_name} : count: {len(remove_vm_tags)} : {remove_vm_tags}')
                else:
                    logger.info(f'Tags are not Removed to the vm: {vm_id} - {vm_name}, something might fail')
            except Exception as err:
                logger.info(f'{err}')
        return remove_vm_tags

    def tag_virtual_machine(self, vm_id: str):
        """
        This method perform the tag operations - read, update, remove
        @param vm_id:
        @return:
        """
        username, vm_name = self.get_virtual_machine_username(vm_id=vm_id)
        vm_tags = self._classic_operations.get_virtual_machine_tags(vm_id=str(vm_id))
        user_tags = self._ibm_client.get_user_tags_from_gsheet(username=username)
        if self._tag_operation == 'remove':
            tags = self.tag_remove_virtual_machine(user_tags=user_tags, vm_tags=vm_tags, vm_id=vm_id, vm_name=vm_name)
        else:
            tags = self.tag_update_virtual_machine(user_tags=user_tags, vm_tags=vm_tags, vm_id=vm_id, vm_name=vm_name)
        return tags

    @logger_time_stamp
    def run(self, vm_id: str = ''):
        """
        This method tag vm ( virtual machines ) from the user tags from the gsheet
        @return:
        """
        response = []
        if vm_id:
            response = self.tag_virtual_machine(vm_id=vm_id)
        else:
            vm_ids = self._classic_operations.get_virtual_machine_ids()
            for vm_id in vm_ids:
                response_data = self.tag_virtual_machine(vm_id=vm_id.get('id'))
                if response_data:
                    response.append({vm_id.get('hostname'): response_data})
        return response



from datetime import datetime, timezone
from dataclasses import dataclass, asdict

from cloud_governance.common.utils.utils import Utils


@dataclass
class PolicyEsMetaData(dict):

    """
    This class is the data modal for policy elasticsearch data
    """

    account: str
    resource_id: str
    user: str
    skip_policy: str
    dry_run: str
    name: str
    region_name: str
    public_cloud: str
    expire_days: int

    unit_price: float = ''
    total_yearly_savings: float = ''
    resource_type: str = ''
    resource_state: str = ''
    clean_up_days: int = ''
    days_count: int = ''
    resource_action: str = ''
    IndexId: str = ''
    SnapshotDate: str = datetime.now(timezone.utc).date().__str__()

    # Specific Policy Attributes
    volume_size: int = ''
    instance_type: str = ''
    launch_time: str = ''
    running_days: int = ''
    create_date: str = ''
    age_days: int = ''
    last_activity_days: int = ''

    def __post_init__(self):
        """
        This method initializes the IndexId
        :return:
        :rtype:
        """
        self.IndexId = (f'{self.SnapshotDate}-{self.public_cloud}-{self.account}-{self.region_name}-{self.resource_id}-'
                        f'{self.resource_state}').lower()

    def get_as_dict_title_case(self):
        """
        This method returns the dict object
        :return:
        :rtype:
        """
        return {Utils.convert_to_title_case(k): v for k, v in asdict(self).items() if v != ''}




from datetime import datetime, timedelta


class ElasticSearchQueries:
    """
    This class having the ElasticSearch Queries used in the cloud-resource-orchestration
    """

    def __init__(self, cro_duration_days: int = 30):
        self.current_end_date = datetime.utcnow()
        self.current_start_date = self.current_end_date - timedelta(days=cro_duration_days)

    def get_all_in_progress_tickets(self, match_conditions: list = None, fields: list = None,  **kwargs):
        """
        This method returns all the in progress tickets
        :param fields:
        :param match_conditions: # [{"term": {"key": "value"}}, {"term": {"key": "value"}}]
        :param kwargs:
        :return:
        """
        if not match_conditions:
            match_conditions = []
        query = \
            {  # check user opened the ticket in elastic_search
                "query": {
                    "bool": {
                        "must": [
                            {"terms": {"ticket_id_state.keyword": ['in-progress']}},
                            *match_conditions,
                        ],
                        "filter": {
                            "range": {
                                "timestamp": {
                                    "format": "yyyy-MM-dd",
                                    "lte": str(self.current_end_date.date()),
                                    "gte": str(self.current_start_date.date()),
                                }
                            }
                        }
                    }
                },
            }
        if fields:
            query['_source'] = fields
        return query

    def get_all_closed_tickets(self, match_conditions: list = None, fields: list = None):
        """
        This method returns the closed tickets
        :return:
        :return:
        """
        if not match_conditions:
            match_conditions = []
        query = \
            {  # check user opened the ticket in elastic_search
                "query": {
                    "bool": {
                        "must": [
                            {"terms": {"ticket_id_state.keyword": ['closed']}},
                            *match_conditions,
                        ],
                        "filter": {
                            "range": {
                                "timestamp": {
                                    "format": "yyyy-MM-dd",
                                    "lte": str(self.current_end_date.date()),
                                    "gte": str(self.current_start_date.date()),
                                }
                            }
                        }
                    }
                },
            }
        if fields:
            query['_source'] = fields
        return query



from ibm_schematics.schematics_v1 import SchematicsV1

from cloud_governance.common.clouds.ibm.account.ibm_authenticator import IBMAuthenticator


class SchematicOperations(IBMAuthenticator):
    """
    This class performs schematic operations.
    """
    REGION_SCHEMATICS_URL = "https://%s.schematics.cloud.ibm.com/"

    def __init__(self):
        super().__init__()
        self.__client = SchematicsV1(self.iam_authenticator)
        self.__client.set_service_url('https://us.schematics.cloud.ibm.com')

    def set_service_url(self, region: str):
        """
        This method sets the service URL.
        :param region:
        :return:
        """
        service_url = self.REGION_SCHEMATICS_URL % region
        self.__client.set_service_url(service_url)

    def get_workspaces(self):
        """
        This method lists all available schematics workspaces
        :return:
        """
        response = self.__client.list_workspaces().get_result()
        return response['workspaces']

    def get_supported_locations(self):
        """
        This method lists supported locations
        :return:
        """
        response = self.__client.list_locations().get_result()
        return response['locations']

    def get_all_workspaces(self):
        """
        This method lists all available schematics workspaces
        :return:
        """
        locations = self.get_supported_locations()
        resources_list = {}
        for location in locations:
            region = location['region']
            geography_code = location['geography_code']
            self.set_service_url(region)
            if geography_code not in resources_list:
                resources_list[geography_code] = self.get_workspaces()
        return resources_list



import logging
from abc import ABC, abstractmethod
from ast import literal_eval
from datetime import datetime, timedelta

import typeguard

from cloud_governance.cloud_resource_orchestration.utils.constant_variables import CRO_OVER_USAGE_ALERT, DATE_FORMAT, \
    OVER_USAGE_THRESHOLD, DEFAULT_ROUND_DIGITS
from cloud_governance.cloud_resource_orchestration.utils.elastic_search_queries import ElasticSearchQueries
from cloud_governance.common.elasticsearch.elasticsearch_operations import ElasticSearchOperations
from cloud_governance.common.ldap.ldap_search import LdapSearch
from cloud_governance.common.logger.init_logger import handler
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp
from cloud_governance.common.mails.mail_message import MailMessage
from cloud_governance.common.mails.postfix import Postfix
from cloud_governance.main.environment_variables import environment_variables


class AbstractCostOverUsage(ABC):

    FORECAST_GRANULARITY = 'MONTHLY'
    FORECAST_COST_METRIC = 'UNBLENDED_COST'

    def __init__(self):
        self._environment_variables_dict = environment_variables.environment_variables_dict
        self._public_cloud_name = self._environment_variables_dict.get('PUBLIC_CLOUD_NAME')
        self._account = self._environment_variables_dict.get('account', '').replace('OPENSHIFT-', '').strip()
        self._es_host = self._environment_variables_dict.get('es_host', '')
        self._es_port = self._environment_variables_dict.get('es_port', '')
        self._over_usage_amount = self._environment_variables_dict.get('CRO_COST_OVER_USAGE', '')
        self._es_ce_reports_index = self._environment_variables_dict.get('USER_COST_INDEX', '')
        self._ldap_search = LdapSearch(ldap_host_name=self._environment_variables_dict.get('LDAP_HOST_NAME', ''))
        self._cro_admins = self._environment_variables_dict.get('CRO_DEFAULT_ADMINS', [])
        self.es_index_cro = self._environment_variables_dict.get('CRO_ES_INDEX', '')
        self._cro_duration_days = self._environment_variables_dict.get('CRO_DURATION_DAYS')
        self._over_usage_threshold = OVER_USAGE_THRESHOLD * self._over_usage_amount
        self.current_end_date = datetime.utcnow()
        self.current_start_date = self.current_end_date - timedelta(days=self._cro_duration_days)
        self.es_operations = ElasticSearchOperations(es_host=self._es_host, es_port=self._es_port)
        self._elastic_search_queries = ElasticSearchQueries(cro_duration_days=self._cro_duration_days)
        self._postfix_mail = Postfix()
        self._mail_message = MailMessage()

    @typeguard.typechecked
    @logger_time_stamp
    def get_monthly_user_es_cost_data(self, tag_name: str = 'User', start_date: datetime = None,
                                      end_date: datetime = None, extra_matches: any = None,
                                      granularity: str = 'MONTHLY', extra_operation: str = 'And', **kwargs):
        """
        This method gets the user cost from the es-data
        :param tag_name: by default User
        :param start_date:
        :param end_date:
        :param extra_matches:
        :param granularity: by default MONTHLY
        :param extra_operation:
        :return:
        """
        start_date, end_date = self.__get_start_end_dates(start_date=start_date, end_date=end_date)
        return self._get_cost_based_on_tag(start_date=str(start_date), end_date=str(end_date), tag_name=tag_name,
                                           granularity=granularity, extra_filters=extra_matches,
                                           extra_operation=extra_operation, **kwargs)

    def _get_forecast_cost_data(self, tag_name: str = 'User', start_date: datetime = None, end_date: datetime = None,
                               extra_matches: any = None, granularity: str = 'MONTHLY', extra_operation: str = 'And'):
        """
        This method returns the forecast based on inputs
        :param tag_name: by default User
        :param start_date:
        :param end_date:
        :param extra_matches:
        :param granularity: by default MONTHLY
        :param extra_operation:
        :return:
        """
        start_date, end_date = self.__get_start_end_dates(start_date=start_date, end_date=end_date)
        return self._get_cost_based_on_tag(start_date=str(start_date), end_date=str(end_date), tag_name=tag_name,
                                           granularity=granularity, extra_filters=extra_matches,
                                           extra_operation=extra_operation, forecast=True)

    @typeguard.typechecked
    @logger_time_stamp
    def _get_user_active_ticket_costs(self, user_name: str):
        """
        This method returns a boolean indicating whether the user should open the ticket or not
        :param user_name:
        :return:
        """
        query = {  # check user opened the ticket in elastic_search
            "query": {
                "bool": {
                    "must": [{"term": {"user_cro.keyword": user_name}},
                             {"terms": {"ticket_id_state.keyword": ['new', 'manager-approved', 'in-progress']}},
                             {"term": {"account_name.keyword": self._account.upper()}},
                             {"term": {"cloud_name.keyword": self._public_cloud_name.upper()}},
                             ],
                    "filter": {
                        "range": {
                            "timestamp": {
                                "format": "yyyy-MM-dd",
                                "lte": str(self.current_end_date.date()),
                                "gte": str(self.current_start_date.date()),
                            }
                        }
                    }
                }
            }
        }
        user_active_tickets = self.es_operations.fetch_data_by_es_query(es_index=self.es_index_cro, query=query)
        if not user_active_tickets:
            return None
        else:
            total_active_ticket_cost = 0
            for cro_data in user_active_tickets:
                opened_ticket_cost = float(cro_data.get('_source').get('estimated_cost'))
                total_active_ticket_cost += opened_ticket_cost
            return total_active_ticket_cost

    @typeguard.typechecked
    @logger_time_stamp
    def _get_user_closed_ticket_costs(self, user_name: str):
        """
        This method returns the users closed tickets cost
        :param user_name:
        :type user_name:
        :return:
        :rtype:
        """
        match_conditions = [{"term": {"user.keyword": user_name}},
                            {"term": {"account_name.keyword": self._account.upper()}},
                            {"term": {"cloud_name.keyword": self._public_cloud_name}}
                            ]
        query = self._elastic_search_queries.get_all_closed_tickets(match_conditions=match_conditions)
        user_closed_tickets = self.es_operations.fetch_data_by_es_query(es_index=self.es_index_cro, query=query,
                                                                        filter_path='hits.hits._source')
        total_closed_ticket_cost = 0
        for closed_ticket in user_closed_tickets:
            total_used_cost = 0
            user_daily_report = closed_ticket.get('_source', {}).get('user_daily_cost', '')
            if user_daily_report:
                user_daily_report = literal_eval(user_daily_report)
                for date, user_cost in user_daily_report.items():
                    if datetime.strptime(date, DATE_FORMAT) >= self.current_start_date:
                        total_used_cost += int(user_cost.get('TicketId', 0))
            total_closed_ticket_cost += total_used_cost
        return total_closed_ticket_cost

    @typeguard.typechecked
    @logger_time_stamp
    def __get_start_end_dates(self, start_date: datetime = None, end_date: datetime = None):
        """
        This method returns the start_date and end_date
        :param start_date:
        :param end_date:
        :return:
        """
        if not start_date:
            start_date = self.current_start_date
        if not end_date:
            end_date = self.current_end_date
        return start_date.date(), end_date.date()

    @logger_time_stamp
    def _get_cost_over_usage_users(self):
        """
        This method returns the cost over usage users which are not opened ticket
        :return:
        """
        over_usage_users = []
        current_month_users = self.get_monthly_user_es_cost_data()
        for user in current_month_users:
            user_name = str(user.get('User'))
            user_cost = round(user.get('Cost'), DEFAULT_ROUND_DIGITS)
            if user_cost >= (self._over_usage_amount - self._over_usage_threshold):
                user_active_tickets_cost = self._get_user_active_ticket_costs(user_name=user_name.lower())
                user_closed_tickets_cost = self._get_user_closed_ticket_costs(user_name=user_name.lower())
                if not user_active_tickets_cost:
                    over_usage_users.append(user)
                else:
                    user_cost_without_active_ticket = user_cost - user_active_tickets_cost - user_closed_tickets_cost
                    if user_cost_without_active_ticket > self._over_usage_amount:
                        user['Cost'] = user_cost_without_active_ticket
                        over_usage_users.append(user)
        return over_usage_users

    @logger_time_stamp
    def _send_alerts_to_over_usage_users(self):
        users_list = self._get_cost_over_usage_users()
        alerted_users = []
        for row in users_list:
            user, cost, project = row.get('User'), row.get('Cost'), row.get('Project', '')
            alerted_users.append(user)
            cc = [*self._cro_admins]
            user_details = self._ldap_search.get_user_details(user_name=user)
            if user_details:
                if self._verify_active_resources(tag_name='User', tag_value=str(user)):
                    name = f'{user_details.get("FullName")}'
                    cc.append(user_details.get('managerId'))
                    subject, body = self._mail_message.cro_cost_over_usage(CloudName=self._public_cloud_name,
                                                                           OverUsageCost=self._over_usage_amount,
                                                                           FullName=name, Cost=cost, Project=project,
                                                                           to=user)
                    es_data = {'Alert': 1, 'MissingUserTicketCost': cost, 'Cloud': self._public_cloud_name}
                    handler.setLevel(logging.WARN)
                    self._postfix_mail.send_email_postfix(to=user, cc=[], content=body, subject=subject,
                                                          mime_type='html', es_data=es_data,
                                                          message_type=CRO_OVER_USAGE_ALERT)
                    handler.setLevel(logging.INFO)
        return alerted_users

    @logger_time_stamp
    def run(self):
        """
        This method runs the cost over usage alerts to users
        :return:
        :rtype:
        """
        return self._send_alerts_to_over_usage_users()

    @abstractmethod
    def _verify_active_resources(self, tag_name: str, tag_value: str) -> bool:
        raise NotImplementedError

    @abstractmethod
    def _get_cost_based_on_tag(self, start_date: str, end_date: str, tag_name: str, extra_filters: any = None,
                               extra_operation: str = 'And', granularity: str = None, forecast: bool = False, **kwargs):
        raise NotImplementedError



from cloud_governance.common.utils.configs import UNUSED_ACCESS_KEY_DAYS, UNUSED_ACCESS_KEY_MAX_DAY
from cloud_governance.policy.helpers.aws.aws_policy_operations import AWSPolicyOperations


class UnusedAccessKey(AWSPolicyOperations):
    RESOURCE_ACTION = "DeActivate"

    def __init__(self):
        super().__init__()

    def run_policy_operations(self):
        """
        This method returns a list of users with at least one active access key whose last used date is greater than UNUSED_ACCESS_KEY_DAYS
        :return:
        :rtype:
        """
        unused_access_keys = []
        iam_users_access_keys = self._get_iam_users_access_keys()
        for username, user_data in iam_users_access_keys.items():
            for access_key_label, access_key_data in user_data.items():
                if 'access key' in access_key_label.lower():
                    last_activity_days = access_key_data['last_activity_days']
                    age_days = access_key_data['age_days']
                    region = user_data['region']
                    user_name = username
                    tags = user_data.get('Tags', [])
                    cleanup_result = False
                    cleanup_days = 0
                    if last_activity_days and int(last_activity_days) >= UNUSED_ACCESS_KEY_DAYS and self._has_active_access_keys(user_name, access_key_label) and self.get_skip_policy_value(tags=tags) not in ('NOTDELETE', 'SKIP'):
                        cleanup_days = self.get_clean_up_days_count(tags=tags)
                        cleanup_result = self.verify_and_delete_resource(resource_id=user_name, tags=tags,
                                                                         clean_up_days=cleanup_days, access_key_label=access_key_label)
                        resource_data = self._get_es_schema(resource_id=user_name,
                                                            user=self.get_tag_name_from_tags(tags=tags, tag_name='User'),
                                                            skip_policy=self.get_skip_policy_value(tags=tags),
                                                            cleanup_days=cleanup_days,
                                                            dry_run=self._dry_run,
                                                            name=user_name,
                                                            region=region,
                                                            cleanup_result=str(cleanup_result),
                                                            resource_action=self.RESOURCE_ACTION,
                                                            cloud_name=self._cloud_name,
                                                            resource_type='UnusedAccessKey',
                                                            resource_state='Active',
                                                            age_days=age_days,
                                                            last_activity_days=last_activity_days,
                                                            unit_price=0)
                        unused_access_keys.append(resource_data)
                    if not cleanup_result:
                        self.update_resource_day_count_tag(resource_id=user_name, cleanup_days=cleanup_days, tags=tags)

        return unused_access_keys



import boto3

from cloud_governance.common.logger.init_logger import logger

# @Todo, This class will be used in the feature releases.
# @Todo, it helps in find the resource data like tags by using search query


class ResourceExplorerOperations:
    """
    This class performs the resource explorer operations
    """

    def __init__(self):
        self.__client = self.__set_client()

    def __set_client(self):
        view = self.list_views()
        region = 'us-east-1'
        if view:
            region = view.split(':')[3]
        return boto3.client('resource-explorer-2', region_name=region)

    def __search(self, search_string: str):
        try:
            response = self.__client.search(QueryString=search_string)
            return response.get('Resources', [])
        except Exception as err:
            logger.error(err)
            return []

    def find_resource_tags(self, resource_id: str):
        search_results = self.__search(search_string=f'"{resource_id}"')
        tags = []
        for resource in search_results:
            if resource_id in resource.get('Arn'):
                if resource.get('Properties'):
                    tags = resource.get('Properties', {})[0].get('Data')
        return tags

    def list_views(self):
        """
        This method returns list the views
        :return:
        :rtype:
        """
        client = boto3.client('resource-explorer-2', region_name='us-east-1')
        views = client.list_views()['Views']
        if views:
            return views[0]
        else:
            raise Exception("No Resource Explorer view found in Region: us-east-1, create one on Free of Charge")



import json
import tempfile
from datetime import datetime

import typeguard

from cloud_governance.cloud_resource_orchestration.clouds.aws.ec2.aws_tagging_operations import AWSTaggingOperations
from cloud_governance.cloud_resource_orchestration.common.abstract_monitor_tickets import AbstractMonitorTickets
from cloud_governance.cloud_resource_orchestration.utils.common_operations import get_tag_value_by_name
from cloud_governance.common.clouds.aws.athena.pyathena_operations import PyAthenaOperations
from cloud_governance.common.clouds.aws.ec2.ec2_operations import EC2Operations
from cloud_governance.common.elasticsearch.elasticsearch_operations import ElasticSearchOperations
from cloud_governance.common.jira.jira_operations import JiraOperations
from cloud_governance.common.ldap.ldap_search import LdapSearch
from cloud_governance.common.logger.init_logger import handler, logger
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp
from cloud_governance.common.mails.mail_message import MailMessage
from cloud_governance.common.mails.postfix import Postfix
from cloud_governance.main.environment_variables import environment_variables


class AWSMonitorTickets(AbstractMonitorTickets):
    """This method monitor the Jira Tickets"""

    NEW = 'New'
    REFINEMENT = 'Refinement'
    CLOSED = 'Closed'
    IN_PROGRESS = 'In Progress'
    CLOSE_JIRA_TICKET = 0
    FIRST_CRO_ALERT: int = 5
    SECOND_CRO_ALERT: int = 3
    DEFAULT_ROUND_DIGITS: int = 3

    def __init__(self, region_name: str = ''):
        super().__init__()
        self.__environment_variables_dict = environment_variables.environment_variables_dict
        self.__cro_resource_tag_name = self.__environment_variables_dict.get('CRO_RESOURCE_TAG_NAME')
        self.__jira_operations = JiraOperations()
        self.__region_name = region_name if region_name else self.__environment_variables_dict.get('AWS_DEFAULT_REGION')
        self.es_cro_index = self.__environment_variables_dict.get('CRO_ES_INDEX', '')
        self.__default_admins = self.__environment_variables_dict.get('CRO_DEFAULT_ADMINS', [])
        self.__cloud_name = self.__environment_variables_dict.get('PUBLIC_CLOUD_NAME', '')
        self.__account = self.__environment_variables_dict.get('account', '')
        self.__es_host = self.__environment_variables_dict.get('es_host', '')
        self.__es_port = self.__environment_variables_dict.get('es_port', '')
        self.__es_operations = ElasticSearchOperations(es_host=self.__es_host, es_port=self.__es_port)
        self.__manager_escalation_days = self.__environment_variables_dict.get('MANAGER_ESCALATION_DAYS')
        self.__ldap_search = LdapSearch(self.__environment_variables_dict.get('LDAP_HOST_NAME'))
        self.__global_admin_name = self.__environment_variables_dict.get('GLOBAL_CLOUD_ADMIN')
        self.__database_name = self.__environment_variables_dict.get('ATHENA_DATABASE_NAME')
        self.__table_name = self.__environment_variables_dict.get('ATHENA_TABLE_NAME')
        self.__athena_account_access_key = self.__environment_variables_dict.get('ATHENA_ACCOUNT_ACCESS_KEY')
        self.__athena_account_secret_key = self.__environment_variables_dict.get('ATHENA_ACCOUNT_SECRET_KEY')
        self.__mail_message = MailMessage()
        self.__postfix = Postfix()
        self.__ec2_operations = EC2Operations()

    @typeguard.typechecked
    @logger_time_stamp
    def get_tickets(self, ticket_status: str):
        """
        This method return the tickets based on status
        :param ticket_status:
        :return:
        """
        return self.__jira_operations.get_all_issues(ticket_status=ticket_status)

    @typeguard.typechecked
    @logger_time_stamp
    def __send_ticket_status_alerts(self, tickets: dict, ticket_status: str):
        """
        This method send alert to user Ticket status if it is on New, Refinement states
        Ticket States:
            New - Need approval
            Refinement - User didn't add the tag to the resources
        :param tickets:
        :param ticket_status:
        :return:
        """
        if ticket_status in (self.NEW, self.REFINEMENT):
            user_tickets = {}
            for ticket_id, description in tickets.items():
                ticket_id = ticket_id.split('-')[-1]
                if self.__account in description.get('AccountName'):
                    if not self.__es_operations.verify_elastic_index_doc_id(index=self.es_cro_index, doc_id=ticket_id):
                        if ticket_status == self.REFINEMENT:
                            ticket_status = 'manager-approved'
                        source = {'cloud_name': description.get('CloudName'),
                                  'account_name': description.get('AccountName').replace('OPENSHIFT-', ''),
                                  'region_name': description.get('Region'), 'user': '',
                                  'user_cro': description.get('EmailAddress').split('@')[0], 'user_cost': 0,
                                  'ticket_id': ticket_id, 'ticket_id_state': ticket_status.lower(),
                                  'estimated_cost': description.get('CostEstimation'), 'instances_count': 0,
                                  'monitored_days': 0,
                                  'ticket_opened_date': description.get('TicketOpenedDate').date(),
                                  'duration': description.get('Days'), 'approved_manager': '',
                                  'user_manager': '', 'project': description.get('Project'),
                                  'owner': f'{description.get("FirstName")} {description.get("LastName")}'.upper(),
                                  'total_spots': 0,
                                  'total_ondemand': 0, 'AllocatedBudget': [], 'instances_list': [],
                                  'instance_types_list': []}
                        self.__es_operations.upload_to_elasticsearch(index=self.es_cro_index, data=source, id=ticket_id)
                    current_date = datetime.now().date()
                    ticket_opened_date = description.get('TicketOpenedDate').date()
                    if ticket_opened_date != current_date:
                        user = description.get('EmailAddress').split('@')[0]
                        manager = description.get('ManagerApprovalAddress').split('@')[0]
                        cc = self.__default_admins
                        subject = body = to = None
                        ticket_opened_days = (current_date - ticket_opened_date).days
                        if ticket_status == self.NEW:  # alert manager if didn't take any action
                            to = manager
                            extra_message = ''
                            if self.__manager_escalation_days <= ticket_opened_days <= self.__manager_escalation_days + 2:
                                manager_of_manager = self.__ldap_search.get_user_details(user_name=manager)
                                to = manager_of_manager.get('ManagerId', '')
                                extra_message = f"Your associate/Manager: [{manager}] doesn't approve this request.<br/>The user {user} is waiting for approval for last {ticket_opened_days} days.<br/>" \
                                                f"Please review the below details and approve/ reject"
                            elif ticket_opened_days >= self.__manager_escalation_days + 2:
                                to = self.__global_admin_name
                                extra_message = f"<b>Missing manager approval.<br />The user {user} is waiting for approval for last {ticket_opened_days} days.<br/>Please review the below details and approve/reject"
                            subject, body = self.__mail_message.cro_request_for_manager_approval(manager=to,
                                                                                                 request_user=user,
                                                                                                 cloud_name=self.__cloud_name,
                                                                                                 ticket_id=ticket_id,
                                                                                                 description=description,
                                                                                                 extra_message=extra_message)
                        else:  # alert user if doesn't add tag name
                            user_tickets.setdefault(user, []).append(f"{ticket_id} : {description.get('Project')}")
            if user_tickets:
                for user, ticket_ids in user_tickets.items():
                    active_instances = self.__ec2_operations.get_active_instances(ignore_tag='TicketId', tag_value=user,
                                                                                  tag_name='User')
                    if active_instances:
                        for region, instances_list in active_instances.items():
                            active_instances_ids = {region: [instance.get('InstanceId') for instance in instances_list]}
                            to = user
                            cc = self.__default_admins
                            subject, body = self.__mail_message.cro_send_user_alert_to_add_tags(user=user,
                                                                                                ticket_ids=ticket_ids)
                            with tempfile.NamedTemporaryFile(mode='w', suffix='.json') as filename:
                                filename.write(json.dumps(active_instances_ids))
                                filename.flush()
                                self.__postfix.send_email_postfix(to=to, cc=cc, subject=subject, content=body,
                                                                  mime_type='html', filename=filename.name)

    @logger_time_stamp
    def __track_tickets(self):
        """
        This method trak the user tickets
        :return:
        """
        self.__send_ticket_status_alerts(ticket_status=self.NEW, tickets=self.get_tickets(ticket_status=self.NEW))
        self.__send_ticket_status_alerts(ticket_status=self.REFINEMENT,
                                         tickets=self.get_tickets(ticket_status=self.REFINEMENT))

    def update_budget_tag_to_resources(self, region_name: str, ticket_id: str, updated_budget: int):
        """
        This method updates the budget to the aws resources which have the tag TicketId: #
        :param region_name:
        :param ticket_id:
        :param updated_budget:
        :return:
        """
        try:
            tag_to_be_updated = 'EstimatedCost'
            tagging_operations = AWSTaggingOperations(region_name=region_name)
            resources_list_to_update = tagging_operations.get_resources_list(tag_name='TicketId', tag_value=ticket_id)
            if resources_list_to_update:
                resource_arn_list = []
                previous_cost = 0
                for resource in resources_list_to_update:
                    resource_arn_list.append(resource.get('ResourceARN'))
                    if previous_cost == 0:
                        previous_cost = get_tag_value_by_name(tags=resource.get('Tags'), tag_name=tag_to_be_updated)
                updated_budget += int(float(previous_cost))
                update_tags_dict = {tag_to_be_updated: str(updated_budget)}
                tagging_operations.tag_resources_list(resources_list=resource_arn_list,
                                                      update_tags_dict=update_tags_dict)
            else:
                logger.info('No AWS resources to update the costs')
        except Exception as err:
            logger.error(err)

    def update_duration_tag_to_resources(self, region_name: str, ticket_id: str, updated_duration: int):
        """
        This method updates the budget to cloud resources
        :param region_name:
        :param ticket_id:
        :param updated_duration:
        :return:
        """
        try:
            tag_to_be_updated = 'Duration'
            tagging_operations = AWSTaggingOperations(region_name=region_name)
            resources_list_to_update = tagging_operations.get_resources_list(tag_name='TicketId', tag_value=ticket_id)
            if resources_list_to_update:
                resource_arn_list = []
                previous_duration = 0
                for resource in resources_list_to_update:
                    resource_arn_list.append(resource.get('ResourceARN'))
                    if previous_duration == 0:
                        previous_duration = get_tag_value_by_name(tags=resource.get('Tags'), tag_name=tag_to_be_updated)
                updated_duration += int(float(previous_duration))
                update_tags_dict = {tag_to_be_updated: str(updated_duration)}
                tagging_operations.tag_resources_list(resources_list=resource_arn_list,
                                                      update_tags_dict=update_tags_dict)
            else:
                logger.info('No AWS resources to update the costs')
        except Exception as err:
            logger.error(err)

    def update_cluster_cost(self):
        """
        This method updates the cluster cost.
        :return:
        :rtype:
        """
        in_progress_tickets = self._get_all_in_progress_tickets()
        for ticket_data in in_progress_tickets:
            source_data = ticket_data.get('_source')
            if source_data:
                ticket_id = source_data.get('ticket_id')
                cluster_names = source_data.get('cluster_names', [])
                if cluster_names:
                    query = self.__prepare_athena_query_for_cluster_cost(names=cluster_names)
                    pyathena_operations = PyAthenaOperations(aws_access_key_id=self.__athena_account_access_key,
                                                             aws_secret_access_key=self.__athena_account_secret_key)
                    response = pyathena_operations.execute_query(query_string=query)
                    if response:
                        cluster_cost = {item.get('ClusterName'): item.get('UnblendedCost') for item in response}
                        self.__es_operations.update_elasticsearch_index(index=self.es_cro_index, id=ticket_id,
                                                                        metadata={
                                                                            'cluster_cost': cluster_cost,
                                                                            'timestamp': datetime.utcnow()
                                                                        })

    def __prepare_athena_query_for_cluster_cost(self, names: list):
        """
        This method returns the athena query to get results
        :param names:
        :type names:
        :return:
        :rtype:
        """
        query = ''
        if names:
            cluster_names = [f""" "resource_tags_user_name" LIKE '%{i}%' """ for i in names]
            case_statement = "CASE"
            for cluster_name in names:
                case_statement += f"""\nWHEN resource_tags_user_name like '%{cluster_name}%' THEN '{cluster_name}' """
            case_statement += "\nELSE resource_tags_user_name END as ClusterName"
            query = f"""
            SELECT
            line_item_usage_account_id as AccountId,
            {case_statement},
            ROUND(SUM(line_item_unblended_cost), 3) as UnblendedCost
            FROM "{self.__database_name}"."{self.__table_name}"
            where {"or ".join(cluster_names)}
            group by 1, 2
            """
        return query

    def notify_ticket_closed(self, ticket_id: str, region_name: str = ''):
        """
        This method deletes the tag of the resources
        :param region_name:
        :param ticket_id:
        :return:
        """
        self.delete_tags_to_resource(ticket_id, region_name)

    def delete_tags_to_resource(self, ticket_id: str, region_name: str):
        """
        This method deletes the TicketId tag of the resources
        :param region_name:
        :param ticket_id:
        :return:
        """
        try:
            tag_to_be_deleted = ['Duration', 'TicketId', 'EstimatedCost', 'ApprovedManager']
            tagging_operations = AWSTaggingOperations(region_name=region_name)
            resources_list_to_update = tagging_operations.get_resources_list(tag_name='TicketId', tag_value=ticket_id)
            if resources_list_to_update:
                resource_arn_list = []
                for resource in resources_list_to_update:
                    resource_arn_list.append(resource.get('ResourceARN'))
                tagging_operations.untag_resources_list(resources_list=resource_arn_list,
                                                        delete_tag_keys_list=tag_to_be_deleted)
            else:
                logger.info('No AWS resources tags were deleted')
        except Exception as err:
            logger.error(err)
            raise err

    @logger_time_stamp
    def run(self):
        """
        This method run all methods of jira tickets monitoring
        :return:
        # """
        self.__track_tickets()
        self.monitor_tickets()



import json
from datetime import datetime
import pandas as pd

from cloud_governance.common.elasticsearch.elasticsearch_operations import ElasticSearchOperations
from cloud_governance.common.logger.logger_time_stamp import logger_time_stamp
from cloud_governance.common.clouds.aws.s3.s3_operations import S3Operations
from cloud_governance.common.clouds.aws.price.price import AWSPrice


class ESUploader:
    """
    This class upload data to elastic search from s3 bucket
    """

    def __init__(self, **kwargs):
        self.__es_host = kwargs.get('es_host')
        self.__es_port = kwargs.get('es_port')
        self.__es_index = kwargs.get('es_index')
        self.__es_doc_type = kwargs.get('es_doc_type')
        self.__es_add_items = kwargs.get('es_add_items')
        self.__bucket_name = kwargs.get('bucket')
        self.__s3_file_name = kwargs.get('s3_file_name')
        self.__logs_bucket_key = kwargs.get('logs_bucket_key')
        self.__region_name = kwargs.get('region')
        self.__policy_name = kwargs.get('policy')
        self.es_operations = ElasticSearchOperations(es_host=self.__es_host, es_port=self.__es_port,
                                                     region=self.__region_name, bucket=self.__bucket_name,
                                                     logs_bucket_key=self.__logs_bucket_key)
        if self.__region_name:
            self.__s3_operation = S3Operations(region_name=self.__region_name, bucket=self.__bucket_name,
                                               logs_bucket_key=self.__logs_bucket_key)
            self.__aws_price = AWSPrice()

    def __get_cluster_cost(self, data, resource, clusters_launch_time, clusters_user):
        """
        This method aggregate cluster cost data
        @param data:
        @param resource:
        @param clusters_user:
        @return:
        """
        # aggregate ec2/ebs cluster cost data
        resource_data = [item.split('|') for item in data['resources_list']]
        df = pd.DataFrame(resource_data)
        # MUST : every fix, change ec2/ebs title
        # cost column: remove space
        df[df.columns[2]] = df[df.columns[2]].str.strip()
        # cost column: change to float
        df[df.columns[2]] = df[df.columns[2]].astype(float).round(3)
        # group by cluster owned column
        cluster_cost = df.groupby(df.columns[-1])[df.columns[2]].sum()
        cluster_cost_results = []
        # cluster
        # title: cluster# | cost($) | user | launch time | cluster owned
        num = 1
        for name, cost in cluster_cost.items():
            if cost > 0:
                if name == '  ':
                    cluster_cost_results.append(f'non cluster | {round(cost, 3)} ')
                    data['non cluster'] = {'name': f'{resource} (non cluster)', 'cost': round(cost, 3)}
                else:
                    cluster_cost_results.append(
                        f" cluster_{num} | {round(cost, 3)} | {clusters_user.get(name.strip())} | {clusters_launch_time.get(name.strip())} | {name.strip()} ")
                    data[f'cluster_{num}'] = {'name': name.strip(), 'cost': round(cost, 3),
                                              'user': clusters_user.get(name.strip()),
                                              'launch_time': clusters_launch_time.get(name.strip())}
                    num += 1
        return cluster_cost_results

    def __get_user_cost(self, data):
        """
        This method aggregate user cost data
        @param data:
        @return:
        """
        # aggregate ec2/ebs cluster cost data
        resource_data = [item.split('|') for item in data['resources_list']]
        df = pd.DataFrame(resource_data)
        # MUST : every fix, change ec2/ebs title
        # cost column: remove space
        df[df.columns[2]] = df[df.columns[2]].str.strip()
        # cost column: change to float
        df[df.columns[2]] = df[df.columns[2]].astype(float).round(3)
        # group by user
        user_cost = df.groupby(df.columns[1])[df.columns[2]].sum()
        user_cost_results = []
        # user
        # title: user# | cost($) | user
        num = 1
        for user, cost in user_cost.items():
            if cost > 0 and user != '  ' and user:
                user_cost_results.append(f"user_{num} | {round(cost, 3)}  | {user.strip()} ")
                data[f'user_{num}'] = {'name': user.strip(), 'cost': round(cost, 3)}
                num += 1
        return user_cost_results

    def upload_last_policy_to_elasticsearch(self, policy: str, index: str, doc_type: str, s3_json_file: str,
                                            es_add_items: dict = None):
        """
        This method is upload json kubernetes cluster data into elasticsearch
        :param policy:
        :param s3_json_file:
        :param index:
        :param doc_type:
        :param es_add_items:
        :return:
        """
        resource = ''
        launch_time_format = ''
        # fetch data from s3 per region/policy
        data = self.__s3_operation.get_last_s3_policy_content(policy=policy, file_name=s3_json_file)
        if data:
            # cluster owned launch time
            clusters_launch_time_dict = {}
            # cluster owned user name
            cluster_user = {}
            data_list = json.loads(data)
            # if json folding in list need to extract it
            if type(data_list) == list:
                # resources_list is a list of items that was triggered by policy
                data_dict = {'resources_list': []}
                for i, item in enumerate(data_list):
                    ec2_ebs_name = ''
                    gitleaks_leakurl = ''
                    user = ''
                    # cluster resource tag
                    cluster_owned = ''
                    # filter all data to save place
                    # data_dict[f'resource{i + 1}'] = item
                    # data_dict[f'resource_{i + 1}'] = 1
                    data_dict['resources'] = i + 1
                    # ec2/ebs
                    if item.get('Tags'):
                        for val in item['Tags']:
                            if val['Key'] == 'Name':
                                ec2_ebs_name = val['Value']
                            if val['Value'] == 'owned':
                                cluster_owned = val['Key']
                            if val['Key'].lower() == 'user':
                                user = val['Value'].lower()
                    if cluster_owned:
                        cluster_user[cluster_owned] = user
                    # ec2 - MUST: every fix, change also cluster title
                    # title:  instance id | user | cost($) | state | instance type | launch time | name | cluster owned
                    if item.get('InstanceId'):
                        resource = 'ec2'
                        # lt_datetime = datetime.strptime(item['LaunchTime'], '%Y-%m-%dT%H:%M:%S+00:00')
                        ec2_cost = self.__aws_price.get_ec2_price(resource=resource, item_data=item)
                        launch_time_format = item['LaunchTime'][:-15].replace('T', ' ')
                        if cluster_owned:
                            clusters_launch_time_dict[cluster_owned] = launch_time_format
                        data_dict['resources_list'].append(
                            f"{item['InstanceId']} | {user} | {ec2_cost} | {item['State']['Name']} | {item['InstanceType']}  | {launch_time_format} | {ec2_ebs_name} | {cluster_owned} ")

                    # ebs - MUST: every fix, change also cluster title
                    # title: volume id | user | cost($/month) | state | volume type | create time | size(gb) | name |  cluster owned
                    if item.get('VolumeId'):
                        resource = 'ebs'
                        ebs_monthly_cost = self.__aws_price.get_ec2_price(resource=resource, item_data=item)
                        create_time_format = item['CreateTime'][:-22].replace('T', ' ')
                        if cluster_owned:
                            clusters_launch_time_dict[cluster_owned] = create_time_format
                        data_dict['resources_list'].append(
                            f"{item['VolumeId']} | {user} | {ebs_monthly_cost} | {item['State']} | {item['VolumeType']} | {create_time_format} | {item['Size']} | {ec2_ebs_name} |  {cluster_owned} ")
                    # gitleaks
                    if item.get('leakURL'):
                        gitleaks_leakurl = item.get('leakURL')
                    if item.get('email'):
                        data_dict['resources_list'].append(f"{item.get('email')} | {gitleaks_leakurl}")

                # get cluster cost data only for ec2 and ebs
                if resource:
                    cluster_cost_results = self.__get_cluster_cost(data=data_dict, resource=resource,
                                                                   clusters_launch_time=clusters_launch_time_dict,
                                                                   clusters_user=cluster_user)
                    data_dict['cluster_cost_data'] = cluster_cost_results
                    user_cost_results = self.__get_user_cost(data=data_dict)
                    data_dict['user_cost_data'] = user_cost_results
                data = data_dict
            elif type(data_list) == dict:
                data = data_list

        # no data for policy
        else:
            data = {'resources': 0}

        # Add items
        for key, value in es_add_items.items():
            data[key] = value

        # utcnow - solve timestamp issue
        data['timestamp'] = datetime.utcnow()  # datetime.now()

        # Upload data to elastic search server
        try:
            if isinstance(data, dict):  # JSON Object
                self.es_operations.upload_to_elasticsearch(index=index, data=data)
            else:  # JSON Array
                for record in data:
                    self.es_operations.upload_to_elasticsearch(index=index, data=record)
            return True
        except Exception:
            raise

    @logger_time_stamp
    def upload_to_es(self, account):
        """
        This method upload data to input ELK
        """
        self.__es_add_items.update({'policy': self.__policy_name, 'region': self.__region_name})
        self.upload_last_policy_to_elasticsearch(policy=self.__policy_name, index=self.__es_index,
                                                 doc_type=self.__es_doc_type, s3_json_file=self.__s3_file_name,
                                                 es_add_items=self.__es_add_items)



import datetime

from cloud_governance.common.utils.configs import UNUSED_DAYS
from cloud_governance.common.utils.utils import Utils
from cloud_governance.policy.helpers.aws.aws_policy_operations import AWSPolicyOperations


class UnUsedNatGateway(AWSPolicyOperations):
    """
    This class sends an alert mail for zombie Nat gateways ( based on vpc routes )
    to the user after 4 days and delete after 7 days.
    """

    NAMESPACE = 'AWS/NATGateway'
    RESOURCE_ACTION = "Delete"

    def __init__(self):
        super().__init__()
        self.__active_cluster_ids = self._get_active_cluster_ids()

    def __check_cloud_watch_logs(self, resource_id: str, days: int = UNUSED_DAYS):
        """
        This method returns weather the NatGateway is used in last input days
        :param resource_id:
        :param days:
        :return:
        """
        if days == 0:
            days = 1
        end_time = datetime.datetime.utcnow()
        start_time = end_time - datetime.timedelta(days=days)
        response = self._cloudwatch.get_metric_data(start_time=start_time, end_time=end_time, resource_id=resource_id,
                                                    resource_type='NatGatewayId', namespace=self.NAMESPACE,
                                                    metric_names={'ActiveConnectionCount': 'Count'},
                                                    statistic='Average')['MetricDataResults'][0]
        for value in response.get('Values', []):
            if value > 0:
                return False
        return True

    def __check_nat_gateway_in_routes(self, nat_gateway_id: str):
        """
        This method check the nat gateway present in the routes or not.
        :param nat_gateway_id:
        :return:
        """
        route_tables = self._ec2_client.describe_route_tables()['RouteTables']
        nat_gateway_found = False
        for route_table in route_tables:
            for route in route_table.get('Routes'):
                if route.get('NatGatewayId') == nat_gateway_id:
                    nat_gateway_found = True
        return nat_gateway_found

    def run_policy_operations(self):
        """
        This method returns the list of unattached volumes
        :return:
        :rtype:
        """
        unit_price = self._resource_pricing.get_nat_gateway_unit_price(region_name=self._region)
        unused_nat_gateways = []
        nat_gateways = self._ec2_operations.get_nat_gateways()
        for nat_gateway in nat_gateways:
            tags = nat_gateway.get('Tags', [])
            resource_id = nat_gateway.get('NatGatewayId')
            cleanup_result = False
            cluster_tag = self._get_cluster_tag(tags=tags)
            cleanup_days = 0
            if (Utils.equal_ignore_case(nat_gateway.get('State'), 'available')
                    and cluster_tag not in self.__active_cluster_ids and
                    self.get_skip_policy_value(tags=tags) not in ('NOTDELETE', 'SKIP')):
                if (not self.__check_nat_gateway_in_routes(nat_gateway_id=resource_id) or
                        self.__check_cloud_watch_logs(resource_id=resource_id)):
                    cleanup_days = self.get_clean_up_days_count(tags=tags)
                    cleanup_result = self.verify_and_delete_resource(resource_id=resource_id, tags=tags,
                                                                     clean_up_days=cleanup_days)
                    resource_data = self._get_es_schema(resource_id=resource_id,
                                                        user=self.get_tag_name_from_tags(tags=tags, tag_name='User'),
                                                        skip_policy=self.get_skip_policy_value(tags=tags),
                                                        cleanup_days=cleanup_days, dry_run=self._dry_run,
                                                        name=self.get_tag_name_from_tags(tags=tags, tag_name='Name'),
                                                        region=self._region,
                                                        cleanup_result=str(cleanup_result),
                                                        resource_action=self.RESOURCE_ACTION,
                                                        cloud_name=self._cloud_name,
                                                        resource_type='NatGateway',
                                                        resource_state=nat_gateway.get('State'),
                                                        unit_price=unit_price
                                                        if not cleanup_result else "Deleted")
                    unused_nat_gateways.append(resource_data)
                    if not cleanup_result:
                        self.update_resource_tags(resource_id=resource_id, tags=tags + self.cost_savings_tag)
            else:
                cleanup_days = 0
                self.delete_resource_tags(resource_id=resource_id, tags=self.cost_savings_tag)

            if not cleanup_result:
                self.update_resource_day_count_tag(resource_id=resource_id, cleanup_days=cleanup_days, tags=tags)

        return unused_nat_gateways



from codecs import open
from os import path
from setuptools import setup, find_packages

__version__ = '1.1.385'

here = path.abspath(path.dirname(__file__))

if path.isfile(path.join(here, 'README.md')):
    with open(path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = f.read()
else:
    long_description = ""

setup(
    name='cloud-governance',
    version=__version__,
    description='Cloud Governance Tool',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Red Hat',
    author_email='ebattat@redhat.com, athiruma@redhat.com',
    url='https://github.com/redhat-performance/cloud-governance',
    license="Apache License 2.0",
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12'
    ],

    zip_safe=False,

    # Find all packages (__init__.py)
    packages=find_packages(include=['cloud_governance', 'cloud_governance.*']),

    install_requires=[
        'aiohttp==3.10.11',  # required by jira
        'attrs==21.4.0',  # readthedocs
        'azure-identity==1.16.1',  # azure identity
        'azure-mgmt-billing==6.0.0',  # azure billing management
        'azure-mgmt-costmanagement==3.0.0',  # azure cost management
        'azure-mgmt-subscription==3.1.1',  # azure subscriptions
        'azure-mgmt-resource==23.0.1',
        'azure-mgmt-compute==30.1.0',
        'azure-mgmt-network==25.0.0',
        'azure-mgmt-monitor==6.0.2',
        'boto3==1.33.1',
        'botocore==1.33.8',
        'elasticsearch-dsl==7.4.0',
        'elasticsearch==7.13.4',  # opensearch 1.2.4 for elasticsearch
        'google-api-python-client==2.57.0',  # google drive
        'google-auth-httplib2==0.1.0',  # google drive
        'google-auth-oauthlib==0.5.2',  # google drive
        'google-cloud-bigquery==3.5.0',  # google cloud cost
        'google-cloud-billing==1.9.1',  # google cloud cost
        'ibm-cloud-sdk-core==3.18.0',
        'ibm-cos-sdk==2.13.6',
        'ibm-platform-services==0.27.0',  # IBM Usage reports
        'ibm-schematics==1.1.0',
        'ibm-vpc==0.21.0',
        'myst-parser==1.0.0',  # readthedocs
        'numpy<=1.26.4',  # opensearch 1.2.4 for elasticsearch
        'oauthlib~=3.1.1',  # required by jira
        'pandas',  # latest: aggregate ec2/ebs cluster data
        'PyAthena[Pandas]==3.0.5',  # AWS Athena package
        'PyGitHub==1.55',  # gitleaks
        'python-ldap==3.4.2',  # prerequisite: sudo dnf install -y python39-devel openldap-devel gcc
        'requests==2.32.2',  # rest api & lambda
        'retry==0.9.2',
        'setuptools',  # Requires for python3.12
        'SoftLayer==6.0.0',  # IBM SoftLayer
        'sphinx-rtd-theme==1.0.0',  # readthedocs
        'sphinx==5.0.0',  # readthedocs
        'typeguard==2.13.3',  # checking types
        'typing==3.7.4.3',
        'urllib3==1.26.19'  # required by jira
    ],

    setup_requires=['pytest', 'pytest-runner', 'wheel', 'coverage'],

    include_package_data=True,

    # dependency_links=[]
)



import json
from typing import Union

from cloud_governance.common.clouds.aws.s3.s3_operations import S3Operations
from cloud_governance.common.utils.json_datetime_encoder import JsonDateTimeEncoder
from cloud_governance.common.logger.init_logger import logger
from cloud_governance.policy.policy_runners.common.abstract_upload import AbstractUpload


class UploadS3(AbstractUpload):

    def __init__(self):
        super().__init__()
        self._s3operations = S3Operations(region_name=self._region)

    def upload(self, data: Union[list, dict]):
        """
        This method upload data to S3
        :param data:
        :type data:
        :return:
        :rtype:
        """
        if self._policy_output:
            data = json.dumps(data, cls=JsonDateTimeEncoder)
            try:
                self._s3operations.save_results_to_s3(policy=self._policy.replace('_', '-'),
                                                      policy_output=self._policy_output, policy_result=data)
                logger.info(f"Uploaded the data s3 Bucket: {self._policy_output}")
            except Exception as err:
                logger.error(f"Exection raised on uploading to s3, {err}")
